import { Remote } from "./Remote";
import { validation } from "../decorators/validation";
import { Device } from "./Device";
import { throwError } from "../utils/Error";
import { logger } from "../Logger";

export enum LoginReturnType {
  "OK" = 0,
  "OK but Voicemeeter Application not launched" = 1,
  "Cannot get client (unexpected)" = -1,
  "Unexpected login (logout was expected before)" = -2,
}

export enum VoicemeeterType {
  Basic = 1,
  Banana = 2,
  Potato = 3,
}

export class Control extends Remote {
  private static buses = {
    Potato: { virtual: 3, physical: 5 },
    Banana: { virtual: 2, physical: 3 },
    Basic: { virtual: 1, physical: 1 },
  };
  private static strips = {
    Potato: { virtual: 3, physical: 5 },
    Banana: { virtual: 2, physical: 3 },
    Basic: { virtual: 1, physical: 2 },
  };
  private static totalBuses = (() => {
    const totalBuses = {} as (typeof Control)["totalBuses"];
    for (const voicemeeterVersion in Control.buses) {
      if (
        Object.prototype.hasOwnProperty.call(Control.buses, voicemeeterVersion)
      ) {
        const element =
          Control.buses[voicemeeterVersion as keyof (typeof Control)["buses"]];
        totalBuses[voicemeeterVersion as keyof (typeof Control)["buses"]] =
          element.physical + element.virtual;
      }
    }
    return totalBuses;
  })() as Record<keyof (typeof Control)["buses"], number>;
  private static totalStrips = (() => {
    const totalStrips = {} as (typeof Control)["totalStrips"];
    for (const voicemeeterVersion in Control.strips) {
      if (
        Object.prototype.hasOwnProperty.call(Control.strips, voicemeeterVersion)
      ) {
        const element =
          Control.strips[
            voicemeeterVersion as keyof (typeof Control)["strips"]
          ];
        totalStrips[voicemeeterVersion as keyof (typeof Control)["strips"]] =
          element.physical + element.virtual;
      }
    }
    return totalStrips;
  })() as Record<keyof (typeof Control)["strips"], number>;

  public loggedIn: boolean = false;
  public isParametersUpdated: boolean = false;
  private interval: NodeJS.Timeout | undefined;

  @validation()
  public get voicemeeterType(): VoicemeeterType {
    const buffer = Buffer.alloc(512);
    this.run("VBVMR_GetVoicemeeterType", buffer);
    return buffer.readInt32LE();
  }

  @validation()
  public get voicemeeterVersion(): number[] {
    const buffer = Buffer.alloc(512);
    this.run("VBVMR_GetVoicemeeterVersion", buffer);
    const version = buffer.readInt32LE();
    return [
      (version & 0xff000000) >> 24,
      (version & 0x00ff0000) >> 16,
      (version & 0x0000ff00) >> 8,
      version & 0x000000ff,
    ];
  }

  @validation()
  public get availableBuses() {
    return {
      total:
        Control.totalBuses[
          VoicemeeterType[this.voicemeeterType] as keyof typeof VoicemeeterType
        ],
      ...Control.buses[
        VoicemeeterType[this.voicemeeterType] as keyof typeof VoicemeeterType
      ],
    };
  }

  @validation()
  public get availableStrips() {
    return {
      total:
        Control.totalStrips[
          VoicemeeterType[this.voicemeeterType] as keyof typeof VoicemeeterType
        ],
      ...Control.strips[
        VoicemeeterType[this.voicemeeterType] as keyof typeof VoicemeeterType
      ],
    };
  }

  @validation()
  public get availableInputDevices(): Device[] {
    const numberOfDevices: number = this.run("VBVMR_Input_GetDeviceNumber");
    const devices: Device[] = [];
    for (let i = 0; i < numberOfDevices; i++) {
      let type, deviceName, hardwareId;
      const result = this.run(
        "VBVMR_Input_GetDeviceDescA",
        i,
        type,
        deviceName,
        hardwareId
      );
      logger.log({}, type, deviceName, hardwareId);

      if (result !== 0) throwError("Error happened on VM side");
    }
    return devices;
  }
  @validation()
  public get availableOutputDevices(): Device[] {
    const numberOfDevices: number = this.run("VBVMR_Output_GetDeviceNumber");
    const devices: Device[] = [];
    for (let i = 0; i < numberOfDevices; i++) {}
    return devices;
  }

  public login() {
    if (this.loggedIn) return;
    const createTimer = () => {
      this.interval = setInterval(() => {
        const result = this.run("VBVMR_IsParametersDirty");
        if (result !== 0 && result !== 1)
          throwError(
            `An error occurred while checking for updated parameters. DLL result code: ${result}`
          );
        const isUpdated = result === 1 ? true : false;
        // if (this.isParametersUpdated !== isUpdated && isUpdated)
        //   console.log(`Updated state`);
        this.isParametersUpdated = isUpdated;
      }, 20);
    };
    if (this.interval === undefined) createTimer();
    else {
      clearInterval(this.interval);
      createTimer();
    }
    const result = this.run("VBVMR_Login") as LoginReturnType;
    if (result !== 0) throwError(`DLL result code: ${result}`);
    this.loggedIn = true;

    return result;
  }

  public logout() {
    if (!this.loggedIn) return;
    if (this.interval !== undefined) {
      clearInterval(this.interval);
      this.interval = undefined;
    }
    const result = this.run("VBVMR_Logout") as number;
    if (result !== 0) throwError(`DLL result code: ${result}`);
    this.loggedIn = false;
    return result;
  }
}
